import subprocess, shlex
import whatap.util.process_util as process_util
import whatap.util.thread_util as thread_util
import whatap.util.logging_util as logging_util
import whatap.pack.smbase_pack as smbase_pack
import whatap.agent.conf.configure as conf
import time, os
from StringIO import StringIO
import kstat
import psutil
from . import CPUStat

def executeCmd(command='getconf CLK_TCK'):
    process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE)
    buf = StringIO()
    while True:
        output = process.stdout.readline()
        if output == '' and process.poll() is not None:
            break
        buf.write(output)

    return buf.getvalue()

tckTime = float(executeCmd())

import multiprocessing

cpucheck = False
cpulist = None
cpucore = 0

def GetCpuInfo():
    global cpucheck
    global cpulist
    global cpucore


    if cpucheck:
        return cpulist, cpucore

    lines = process_util.executeCommandShellWithTimeout('''psrinfo | grep on-line''')

    if lines == "":
        logging_util.error("psrinfo command error")
        cpucore = int(multiprocessing.cpu_count())
        cpulist = range(cpucore)
        cpucheck = True
        return cpulist, cpucore

    cpulist = []
    for line in lines.splitlines():
        parts = line.split()
        if len(parts) < 2:
            continue

        try:
            cpu_id = int(parts[0])
            cpulist.append(cpu_id)
        except ValueError:
            logging_util.error("psrinfo cpu id error")
        cpucore = len(cpulist)
        cpucheck = True
        logging_util.info("CPU List : ", ', '.join(map(str, cpulist)))
    return cpulist, cpucore


pagesize = os.sysconf('SC_PAGE_SIZE')
jiff = 100



oldCpu = None
# infra 
def GetCpu():
    global jiff
    cpu_times = psutil.cpu_times(percpu=False)

    global oldCpu
    if not oldCpu:
        oldCpu = cpu_times
        return None

    c = smbase_pack.CpuLinux()
    all_cpu = cpu_times.user + cpu_times.iowait + cpu_times.idle + cpu_times.system
    last_all_cpu = oldCpu.user + oldCpu.iowait + oldCpu.idle + oldCpu.system
    cpu_diff = float(all_cpu - last_all_cpu)
    c.user = jiff * float(cpu_times.user - oldCpu.user) / cpu_diff
    c.system = jiff * float(cpu_times.system - oldCpu.system) / cpu_diff
    c.iowait = jiff * float(cpu_times.iowait - oldCpu.iowait) / cpu_diff
    c.idle = jiff * float(cpu_times.idle - oldCpu.idle) / cpu_diff
    c.nice = float(0)
    c.irq = float(0)
    c.softirq = float(0)
    c.steal = float(0)
    c.load1, c.load5, c.load15= os.getloadavg()
    oldCpu = cpu_times

    return c

oldCpuList = None
def GetCpuCore():
    global jiff
    cpu_times = psutil.cpu_times(percpu=True)

    global oldCpuList
    if not oldCpuList:
        oldCpuList = cpu_times
        return None

    cpuPerf = []
    for index, ctimes in enumerate(cpu_times):
        c = smbase_pack.CpuLinux()
        all_cpu = ctimes.user + ctimes.iowait + ctimes.idle + ctimes.system
        last_all_cpu = oldCpuList[index].user + oldCpuList[index].iowait + oldCpuList[index].idle + oldCpuList[index].system
        cpu_diff = float(all_cpu - last_all_cpu)
        if cpu_diff == 0:
            c.user = 0
            c.system = 0
            c.iowait = 0
            c.idle = 0
        else:
            c.user = jiff * float(ctimes.user - oldCpuList[index].user) / cpu_diff
            c.system = jiff * float(ctimes.system - oldCpuList[index].system) / cpu_diff
            c.iowait = jiff * float(ctimes.iowait - oldCpuList[index].iowait) / cpu_diff
            c.idle = jiff * float(ctimes.idle - oldCpuList[index].idle) / cpu_diff
        c.nice = float(0)
        c.irq = float(0)
        c.softirq = float(0)
        c.steal = float(0)

        cpuPerf.append(c) 

    oldCpuList = cpu_times

    return cpuPerf

oldStats = None 
def GetCpuStat():
    now = time.time()

    cpuStats = psutil.cpu_stats()
    totalfork = 0
    totalexec = 0
    totalrunq = 0
    totalblockq = 0
    totalwaitq = 0

    global k
    if not k:
        k = kstat.Kstat()

    cpulist, cpucore = GetCpuInfo()
    for i in cpulist:
        try:
            km = k['cpu', i, 'sys']
            totalfork += km['sysfork']
            totalexec += km['sysexec']
        except KeyError as e:
            logging_util.error("cpu sys key error", e)
        except Exception as e:
            logging_util.error("cpu sys exception", e)

    km = k['unix', 0, 'sysinfo']
    totalrunq = km['runque']
    totalblockq = km['swpque']
    totalwaitq = km['waiting']

    kernelStats = (totalfork, totalexec, totalrunq, totalblockq, totalwaitq)


    returnFlag = False

    global oldStats
    if not oldStats:
        oldStats = [cpuStats, kernelStats, now]
        return None


    difftime = float(now - oldStats[2])
    oldCpuStats = oldStats[0]
    oldKernelStats = oldStats[1]

    cs = CPUStat() 
    cs.ctx = float(cpuStats.ctx_switches - oldCpuStats.ctx_switches) / difftime
    cs.interrupt = float(cpuStats.interrupts - oldCpuStats.interrupts) / difftime
    cs.syscall = float(cpuStats.syscalls - oldCpuStats.syscalls) / difftime

    cs.forkCount = float(kernelStats[0] - oldKernelStats[0]) / difftime
    cs.execCount = float(kernelStats[1] - oldKernelStats[1]) / difftime
    cs.runqueue = float(kernelStats[2] - oldKernelStats[2]) / difftime
    cs.blockqueue = float(kernelStats[3] - oldKernelStats[3]) / difftime
    cs.waitqueue = float(kernelStats[4] - oldKernelStats[4]) / difftime
    cs.pendingqueue = CPUStat.UNSUPPORTED

    oldStats = [cpuStats, kernelStats, now]

    return cs

def GetSwapMemory():
    m = smbase_pack.MemoryLinux()
    swapmem = psutil.swap_memory()
    m.swapUsed = swapmem.used
    m.swapTotal = swapmem.total

    if m.swapTotal > 0:
        m.swapPused = swapmem.percent
    else:
        m.swapPused = 0

    return m

oldMemoryStat = None
def GetMemory():
    global k
    now = time.time()
    m = smbase_pack.MemoryLinux()
    if conf.GetConfig().SunosMdbEnabled:
        memstat = getMemStat()
        m.total = memstat['total']
        m.free = memstat['free_freelist'] 
        m.cached = memstat['zfs_file_data'] + memstat['free_cachelist']
        if 'page_cache' in memstat:
            m.cached += memstat['page_cache']
        m.available = m.free + m.cached
        m.pavailable = float(m.available) / float(m.total) * float(100.0)
        m.used = m.total - m.available
        m.pused = float(float(m.used)/float(m.total)*100.0)
    else:
        if k == None:
            k = kstat.Kstat()
        kstat_now = k['unix', 0, 'system_pages']
        m.total = kstat_now['pagestotal'] * pagesize
        m.free = kstat_now['freemem'] * pagesize
        try:
            zfsStats = k['zfs', 0, 'arcstats']
            m.cached = zfsStats['size']
        except KeyError:
            m.cached = 0
        except TypeError:
            m.cached = 0
        m.available = m.free + m.cached
        m.pavailable = float(m.available) / float(m.total) * float(100.0)
        m.used = m.total - m.available
        m.pused = float(float(m.used)/float(m.total)*100.0)

    m.buffers = 0
    m.shared = 0
    m.pageFault = 0
    
    swapIn = 0
    swapOut = 0
    fileSystemIn = 0
    fileSystemOut = 0
    scannedPage = 0
    freedPage = 0

    cpulist, cpucore = GetCpuInfo()
    for i in cpulist:
        try:
            km = k['cpu', i, 'vm']
            swapIn += km['pgswapin']
            swapOut += km['pgswapout']
            fileSystemIn += km['fspgin']
            fileSystemOut += km['fspgout']
            scannedPage += km['scan']
            freedPage += km['dfree']
        except KeyError as e:
            logging_util.error("cpu vm key error", e)
        except Exception as e:
            logging_util.error("cpu vm exception", e)
    
    memPageStat = (swapIn, swapOut, fileSystemIn, fileSystemOut, scannedPage, freedPage, now)
    global oldMemoryStat
    if not oldMemoryStat:
        m.swapIn = 0
        m.swapOut = 0
        m.fileSystemIn = 0
        m.fileSystemOut = 0
        m.scannedPage = 0
        m.freedPage = 0
    else:
        difftime = float(now - oldMemoryStat[6])
        m.swapIn = float(memPageStat[0] - oldMemoryStat[0]) / difftime
        m.swapOut = float(memPageStat[1] - oldMemoryStat[1]) / difftime
        m.fileSystemIn = float(memPageStat[2] - oldMemoryStat[2]) / difftime
        m.fileSystemOut = float(memPageStat[3] - oldMemoryStat[3]) / difftime
        m.scannedPage = float(memPageStat[4] - oldMemoryStat[4]) / difftime
        m.freedPage = float(memPageStat[5] - oldMemoryStat[5]) / difftime

    m.buffers = 0
    m.shared = 0
    m.pageFault = 0

    oldMemoryStat = memPageStat
    
    return m

oldVmstat = None
k = None
def GetCpuMemory():
    global jiff

    cpu_times = psutil.cpu_times(percpu=False)
    global oldVmstat
    if not oldVmstat:
        oldVmstat = cpu_times
        return None, None, None

    c = smbase_pack.CpuLinux()
    all_cpu = cpu_times.user + cpu_times.iowait + cpu_times.idle + cpu_times.system
    last_all_cpu = oldVmstat.user + oldVmstat.iowait + oldVmstat.idle + oldVmstat.system
    cpu_diff = float(all_cpu - last_all_cpu)
    c.user = jiff * float(cpu_times.user - oldVmstat.user) / cpu_diff
    c.system = jiff * float(cpu_times.system - oldVmstat.system) / cpu_diff
    c.iowait = jiff * float(cpu_times.iowait - oldVmstat.iowait) / cpu_diff
    c.idle = jiff * float(cpu_times.idle - oldVmstat.idle) / cpu_diff
    c.nice = float(0)
    c.irq = float(0)
    c.softirq = float(0)
    c.steal = float(0)
    c.load1, c.load5, c.load15= os.getloadavg()
    oldVmstat = cpu_times

    p = smbase_pack.MemoryLinux()
    if conf.GetConfig().SunosMdbEnabled:
        memstat = getMemStat()
        p.total = memstat['total']
        p.free = memstat['free_freelist'] 
        p.cached = memstat['zfs_file_data'] + memstat['free_cachelist']
        if 'page_cache' in memstat:
            p.cached += memstat['page_cache']
        p.available = p.free + p.cached
        p.pavailable = float(p.available) / float(p.total) * float(100.0)
        p.used = p.total - p.available
        p.pused = float(float(p.used)/float(p.total)*100.0)
        
    else:
        global k
        if k == None:
            k = kstat.Kstat()
        kstat_now = k['unix', 0, 'system_pages']
        p.total = kstat_now['pagestotal'] * pagesize
        p.free = kstat_now['freemem'] * pagesize
        try:
            zfsStats = k['zfs', 0, 'arcstats']
            p.cached = zfsStats['size']
        except KeyError:
            p.cached = 0
        except TypeError:
            p.cached = 0
        p.available = p.free + p.cached
        p.pavailable = float(p.available) / float(p.total) * float(100.0)
        p.used = p.total - p.available
        p.pused = float(float(p.used)/float(p.total)*100.0)
        
    p.buffers = 0
    p.shared = 0

    swapmem = psutil.swap_memory()
    p.swapUsed = swapmem.used
    p.swapTotal = swapmem.total
    if p.swapTotal > 0:
        p.swapPused = swapmem.percent
    else:
        p.swapPused = 0

    p.pageFault = 0

    uptime= int(time.time()) - int(psutil.boot_time())
    #print 'cpu_memory', int(time.time()) , int(vmstat['boot time']/1024.0), uptime
    return c, p, uptime

"""
Solraris Old
Page Summary                 Pages             Bytes  %Tot
----------------- ----------------  ----------------  ----
Kernel                      162541              1.2G   32%
Guest                            0                 0    0%
ZFS Metadata                 31639            247.1M    6%
ZFS File Data                33225            259.5M    7%
Anon                         64650            505.0M   13%
Exec and libs                 1323             10.3M    0%
Page cache                    7006             54.7M    1%
Free (cachelist)                 1                8k    0%
Free (freelist)             207519              1.5G   41%
Total                       507904              3.8G
"""



"""
Usage Type/Subtype                      Pages    Bytes  %Tot  %Tot/%Subt
---------------------------- ---------------- -------- ----- -----------
Kernel                                 154062     1.1g 14.9%
ZFS                                    389508     2.9g 37.7%
User/Anon                               55774   435.7m  5.4%
Exec and libs                            1287    10.0m  0.1%
Page Cache                               6844    53.4m  0.6%
Free (cachelist)                        20852   162.9m  2.0%
Free                                   403865     3.0g 39.1%
Total                                 1032192     7.8g  100%

"""


memstat_keys = {"Kernel":"kernel",
    "Guest":"guest",
    "ZFS Metadata":"zfs_metadata",
    "ZFS File Data":"zfs_file_data",
    "Anon":"anon",
    "Exec and libs":"exec_and_libs",
    "Page cache":"page_cache",
    "Free (cachelist)":"free_cachelist",
    "Free (freelist)":"free_freelist",
    "Total":"total",
    "ZFS":"zfs_file_data",
    "User/Anon":"anon",
    "Free":"free_freelist",
}


memstats_required_keys = ["total", "free_cachelist", 'free_freelist', 'zfs_file_data']

def parseMemstat(output):
    global memstat_keys
    global pagesize

    if not output:
        return None, None
    for k, v in memstat_keys.items():
        if output.startswith(k):
            line = output.strip(k)
            line = line.strip()
            page = line.split()[0]
            try:
                return v, int(page) * pagesize
            except:
                continue
    return None, None

g_memstat = {"kernel":0,"guest":0,"zfs_metadata":0,"zfs_file_data":0,"anon":0, "exec_and_libs":0, "page_cache":0, "free_cachelist":0, "free_freelist":0, "total":1}

def getMemStat():
    global g_memstat

    return g_memstat

def _getMemStat(command="mdb -k", parser = parseMemstat):
    started=time.time()*1000
    process = subprocess.Popen(shlex.split(command), stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    memoryInfo = {'created': started}
    process.stdin.write('::memstat\n')
    process.stdin.write('::quit\n')
    process.stdin.close()
    endCheck = False
    
    while process.poll() == None:
        output = process.stdout.readline().strip()
        if not output:
            if not endCheck:
                endCheck = True
                continue
            break
        if output:
            #print('getVmStat', output)
            k, v = parser(output)
            if k != None and v != None:
                memoryInfo[k]= v
    logging_util.debug(command, process.stderr.read())
    
    try:
        process.kill()
    except:
        pass

    return memoryInfo

def updateMemstat():
    global g_memstat
    global memstat_keys
    if conf.GetConfig().SunosMdbEnabled:
        memstat_thistime = _getMemStat()
        for required in memstats_required_keys:
            if required not in memstat_thistime:
                return g_memstat
        g_memstat = memstat_thistime
    return g_memstat

try:
    updateMemstat()
    thread_util.asyncLoop(updateMemstat, conf.GetConfig().MemstatInterval)
except Exception, e:
    print(e)

def getVmStat(command="vmstat -s"):
    started=time.time()*1000
    process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE)
    memoryInfo = {'created': started}
    while True:
        output = process.stdout.readline().strip()
        #print 'getVmStat', output
        if not output and process.poll() is not None:
            break
        if output:
            #print 'getVmStat', output
            tokens = output.split()
            v = tokens[0]
            key = ' '.join(tokens[1:])
            memoryInfo[key]= int(v)*1024

    return memoryInfo

def test():
    while True:
        c, p, t = GetCpuMemory()
        if p:
            print('pavailable:', p.pavailable)
        
        time.sleep(5)
    

if __name__ == '__main__':
    test()



"""Test correct operation of the print function.
"""

# In 2.6, this gives us the behavior we want.  In 3.0, it has
#  no function, but it still must parse correctly.
from __future__ import print_function

import unittest
from test import test_support

from StringIO import StringIO

NotDefined = object()

# A dispatch table all 8 combinations of providing
#  sep, end, and file
# I use this machinery so that I'm not just passing default
#  values to print, I'm either passing or not passing in the
#  arguments
dispatch = {
    (False, False, False):
     lambda args, sep, end, file: print(*args),
    (False, False, True):
     lambda args, sep, end, file: print(file=file, *args),
    (False, True,  False):
     lambda args, sep, end, file: print(end=end, *args),
    (False, True,  True):
     lambda args, sep, end, file: print(end=end, file=file, *args),
    (True,  False, False):
     lambda args, sep, end, file: print(sep=sep, *args),
    (True,  False, True):
     lambda args, sep, end, file: print(sep=sep, file=file, *args),
    (True,  True,  False):
     lambda args, sep, end, file: print(sep=sep, end=end, *args),
    (True,  True,  True):
     lambda args, sep, end, file: print(sep=sep, end=end, file=file, *args),
    }

# Class used to test __str__ and print
class ClassWith__str__:
    def __init__(self, x):
        self.x = x
    def __str__(self):
        return self.x

class TestPrint(unittest.TestCase):
    def check(self, expected, args,
            sep=NotDefined, end=NotDefined, file=NotDefined):
        # Capture sys.stdout in a StringIO.  Call print with args,
        #  and with sep, end, and file, if they're defined.  Result
        #  must match expected.

        # Look up the actual function to call, based on if sep, end, and file
        #  are defined
        fn = dispatch[(sep is not NotDefined,
                       end is not NotDefined,
                       file is not NotDefined)]

        with test_support.captured_stdout() as t:
            fn(args, sep, end, file)

        self.assertEqual(t.getvalue(), expected)

    def test_print(self):
        def x(expected, args, sep=NotDefined, end=NotDefined):
            # Run the test 2 ways: not using file, and using
            #  file directed to a StringIO

            self.check(expected, args, sep=sep, end=end)

            # When writing to a file, stdout is expected to be empty
            o = StringIO()
            self.check('', args, sep=sep, end=end, file=o)

            # And o will contain the expected output
            self.assertEqual(o.getvalue(), expected)

        x('\n', ())
        x('a\n', ('a',))
        x('None\n', (None,))
        x('1 2\n', (1, 2))
        x('1   2\n', (1, ' ', 2))
        x('1*2\n', (1, 2), sep='*')
        x('1 s', (1, 's'), end='')
        x('a\nb\n', ('a', 'b'), sep='\n')
        x('1.01', (1.0, 1), sep='', end='')
        x('1*a*1.3+', (1, 'a', 1.3), sep='*', end='+')
        x('a\n\nb\n', ('a\n', 'b'), sep='\n')
        x('\0+ +\0\n', ('\0', ' ', '\0'), sep='+')

        x('a\n b\n', ('a\n', 'b'))
        x('a\n b\n', ('a\n', 'b'), sep=None)
        x('a\n b\n', ('a\n', 'b'), end=None)
        x('a\n b\n', ('a\n', 'b'), sep=None, end=None)

        x('*\n', (ClassWith__str__('*'),))
        x('abc 1\n', (ClassWith__str__('abc'), 1))

        # 2.x unicode tests
        x(u'1 2\n', ('1', u'2'))
        x(u'u\1234\n', (u'u\1234',))
        x(u'  abc 1\n', (' ', ClassWith__str__(u'abc'), 1))

        # errors
        self.assertRaises(TypeError, print, '', sep=3)
        self.assertRaises(TypeError, print, '', end=3)
        self.assertRaises(AttributeError, print, '', file='')

    def test_mixed_args(self):
        # If an unicode arg is passed, sep and end should be unicode, too.
        class Recorder(object):

            def __init__(self, must_be_unicode):
                self.buf = []
                self.force_unicode = must_be_unicode

            def write(self, what):
                if self.force_unicode and not isinstance(what, unicode):
                    raise AssertionError("{0!r} is not unicode".format(what))
                self.buf.append(what)

        buf = Recorder(True)
        print(u'hi', file=buf)
        self.assertEqual(u''.join(buf.buf), 'hi\n')
        del buf.buf[:]
        print(u'hi', u'nothing', file=buf)
        self.assertEqual(u''.join(buf.buf), 'hi nothing\n')
        buf = Recorder(False)
        print('hi', 'bye', end=u'\n', file=buf)
        self.assertIsInstance(buf.buf[1], unicode)
        self.assertIsInstance(buf.buf[3], unicode)
        del buf.buf[:]
        print(sep=u'x', file=buf)
        self.assertIsInstance(buf.buf[-1], unicode)


def test_main():
    test_support.run_unittest(TestPrint)

if __name__ == "__main__":
    test_main()



import whatap.util.process_util as process_util
import whatap.util.logging_util as logging_util

STATE_MAP = {
    'ESTABLISHED': 'netstat_Established',
    'SYN_SENT': 'netstat_Syn_Sent',
    'SYN_RCVD': 'netstat_Syn_Recv',
    'SYN_RECEIVED': 'netstat_Syn_Recv',
    'FIN_WAIT_1': 'netstat_Fin_Wait1',
    'FIN_WAIT1': 'netstat_Fin_Wait1',
    'FIN_WAIT_2': 'netstat_Fin_Wait2',
    'FIN_WAIT2': 'netstat_Fin_Wait2',
    'TIME_WAIT': 'netstat_Time_Wait',
    'CLOSE_WAIT': 'netstat_Close_Wait',
    'LAST_ACK': 'netstat_Last_Ack',
    'LISTEN': 'netstat_Listen',
    'CLOSING': 'netstat_Closing',
    'CLOSE': 'netstat_Unknown',
    'CLOSED': 'netstat_Unknown',
    'IDLE': 'netstat_Unknown',
    'BOUND': 'netstat_Unknown',
}

def GetNetstatCounters():
    counters = {
        'netstat_Unknown': 0,
        'netstat_Established': 0,
        'netstat_Syn_Sent': 0,
        'netstat_Syn_Recv': 0,
        'netstat_Fin_Wait1': 0,
        'netstat_Fin_Wait2': 0,
        'netstat_Time_Wait': 0,
        'netstat_Close_Wait': 0,
        'netstat_Last_Ack': 0,
        'netstat_Listen': 0,
        'netstat_Closing': 0,
        'Server_Filedescriptors': 0,
    }
    psutil_ok = False
    try:
        import psutil as _psutil
        for conn in _psutil.net_connections(kind='tcp'):
            status = conn.status.upper() if conn.status else ''
            key = STATE_MAP.get(status, 'netstat_Unknown')
            counters[key] += 1
        psutil_ok = True
    except Exception:
        psutil_ok = False

    if not psutil_ok:
        try:
            output = process_util.executeCommandShellWithTimeout('netstat -an -f inet 2>/dev/null', timeout=10)
            for line in output.split('\n'):
                tokens = line.strip().split()
                if not tokens or tokens[0] != 'TCP':
                    continue
                if len(tokens) < 6:
                    continue
                state = tokens[-1].upper()
                key = STATE_MAP.get(state, 'netstat_Unknown')
                counters[key] += 1
        except Exception as e:
            logging_util.error('GetNetstatCounters(Solaris) error:', e)

    try:
        import os, glob
        fd_total = 0
        for pid_dir in glob.glob('/proc/*/fd'):
            try:
                fd_total += len(os.listdir(pid_dir))
            except Exception:
                pass
        counters['Server_Filedescriptors'] = fd_total
    except Exception:
        pass

    return counters





class Data:
    def __init__(self):
        self.data = 0 

class Test:
    def __init__(self):
        self.test = Data()

    def run(self):
        self.test = Data()
        self.test.data = 1
        check = self.test
        print (id(self.test))
        print (type(self.test))

        self.test = Data()
        self.test.data = 2
        print (id(self.test))
        print (type(self.test))


        print (check)
        print (check.data)
        print (self.test)
        print (self.test.data)

Test().run()






class ValueEnum(object):
    NULL = 0
    BOOLEAN = 10
    DECIMAL = 20
    FLOAT = 30
    DOUBLE = 40

    DOUBLE_SUMMARY = 45
    LONG_SUMMARY = 46

    TEXT = 50
    TEXT_HASH = 51
    BLOB = 60
    IP4ADDR = 61

    LIST = 70

    ARRAY_INT = 71
    ARRAY_FLOAT = 72
    ARRAY_TEXT = 73
    ARRAY_LONG = 74

    MAP = 80

    @staticmethod
    def create(code):
        from whatap.value.blob_value import BlobValue
        from whatap.value.boolean_value import BooleanValue
        from whatap.value.decimal_value import DecimalValue
        from whatap.value.double_summary import DoubleSummary
        from whatap.value.double_value import DoubleValue
        from whatap.value.float_array import FloatArray
        from whatap.value.float_value import FloatValue
        from whatap.value.int_array import IntArray
        from whatap.value.ip4_value import IP4Value
        from whatap.value.list_value import ListValue
        from whatap.value.long_array import LongArray
        from whatap.value.long_summary import LongSummary
        from whatap.value.map_value import MapValue
        from whatap.value.null_value import NullValue
        from whatap.value.text_array import TextArray
        from whatap.value.text_hash_value import TextHashValue
        from whatap.value.text_value import TextValue

        if code is None:
            return NullValue()
        elif code == ValueEnum.BOOLEAN:
            return BooleanValue()
        elif code == ValueEnum.DECIMAL:
            return DecimalValue()
        elif code == ValueEnum.FLOAT:
            return FloatValue()
        elif code == ValueEnum.DOUBLE:
            return DoubleValue()
        elif code == ValueEnum.TEXT:
            return TextValue()
        elif code == ValueEnum.TEXT_HASH:
            return TextHashValue()
        elif code == ValueEnum.BLOB:
            return BlobValue()
        elif code == ValueEnum.IP4ADDR:
            return IP4Value()
        elif code == ValueEnum.LIST:
            return ListValue()
        elif code == ValueEnum.MAP:
            return MapValue()
        elif code == ValueEnum.LONG_SUMMARY:
            return LongSummary()
        elif code == ValueEnum.DOUBLE_SUMMARY:
            return DoubleSummary()
        elif code == ValueEnum.ARRAY_INT:
            return IntArray()
        elif code == ValueEnum.ARRAY_FLOAT:
            return FloatArray()
        elif code == ValueEnum.ARRAY_TEXT:
            return TextArray()
        elif code == ValueEnum.ARRAY_LONG:
            return LongArray()
        else:
            return None



#----------------------------------------------------------------------
#  Copyright (c) 1999-2001, Digital Creations, Fredericksburg, VA, USA
#  and Andrew Kuchling. All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#
#    o Redistributions of source code must retain the above copyright
#      notice, this list of conditions, and the disclaimer that follows.
#
#    o Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions, and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#
#    o Neither the name of Digital Creations nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS AND CONTRIBUTORS *AS
#  IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
#  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL
#  CREATIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
#  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
#  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
#  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
#  DAMAGE.
#----------------------------------------------------------------------


# This module is just a placeholder for possible future expansion, in
# case we ever want to augment the stuff in _db in any way.  For now
# it just simply imports everything from _db.

import sys
absolute_import = (sys.version_info[0] >= 3)

if not absolute_import :
    if __name__.startswith('bsddb3.') :
        # import _pybsddb binary as it should be the more recent version from
        # a standalone pybsddb addon package than the version included with
        # python as bsddb._bsddb.
        from _pybsddb import *
        from _pybsddb import __version__
    else:
        from _bsddb import *
        from _bsddb import __version__
else :
    # Because this syntaxis is not valid before Python 2.5
    if __name__.startswith('bsddb3.') :
        exec("from ._pybsddb import *")
        exec("from ._pybsddb import __version__")
    else :
        exec("from ._bsddb import *")
        exec("from ._bsddb import __version__")



# XXX TypeErrors on calling handlers, or on bad return values from a
# handler, are obscure and unhelpful.

import StringIO, sys
import unittest

from xml.parsers import expat

from test import test_support
from test.test_support import sortdict, run_unittest


class SetAttributeTest(unittest.TestCase):
    def setUp(self):
        self.parser = expat.ParserCreate(namespace_separator='!')
        self.set_get_pairs = [
            [0, 0],
            [1, 1],
            [2, 1],
            [0, 0],
            ]

    def test_returns_unicode(self):
        for x, y in self.set_get_pairs:
            self.parser.returns_unicode = x
            self.assertEqual(self.parser.returns_unicode, y)

    def test_ordered_attributes(self):
        for x, y in self.set_get_pairs:
            self.parser.ordered_attributes = x
            self.assertEqual(self.parser.ordered_attributes, y)

    def test_specified_attributes(self):
        for x, y in self.set_get_pairs:
            self.parser.specified_attributes = x
            self.assertEqual(self.parser.specified_attributes, y)


data = '''\
<?xml version="1.0" encoding="iso-8859-1" standalone="no"?>
<?xml-stylesheet href="stylesheet.css"?>
<!-- comment data -->
<!DOCTYPE quotations SYSTEM "quotations.dtd" [
<!ELEMENT root ANY>
<!NOTATION notation SYSTEM "notation.jpeg">
<!ENTITY acirc "&#226;">
<!ENTITY external_entity SYSTEM "entity.file">
<!ENTITY unparsed_entity SYSTEM "entity.file" NDATA notation>
%unparsed_entity;
]>

<root attr1="value1" attr2="value2&#8000;">
<myns:subelement xmlns:myns="http://www.python.org/namespace">
     Contents of subelements
</myns:subelement>
<sub2><![CDATA[contents of CDATA section]]></sub2>
&external_entity;
</root>
'''


# Produce UTF-8 output
class ParseTest(unittest.TestCase):
    class Outputter:
        def __init__(self):
            self.out = []

        def StartElementHandler(self, name, attrs):
            self.out.append('Start element: ' + repr(name) + ' ' +
                            sortdict(attrs))

        def EndElementHandler(self, name):
            self.out.append('End element: ' + repr(name))

        def CharacterDataHandler(self, data):
            data = data.strip()
            if data:
                self.out.append('Character data: ' + repr(data))

        def ProcessingInstructionHandler(self, target, data):
            self.out.append('PI: ' + repr(target) + ' ' + repr(data))

        def StartNamespaceDeclHandler(self, prefix, uri):
            self.out.append('NS decl: ' + repr(prefix) + ' ' + repr(uri))

        def EndNamespaceDeclHandler(self, prefix):
            self.out.append('End of NS decl: ' + repr(prefix))

        def StartCdataSectionHandler(self):
            self.out.append('Start of CDATA section')

        def EndCdataSectionHandler(self):
            self.out.append('End of CDATA section')

        def CommentHandler(self, text):
            self.out.append('Comment: ' + repr(text))

        def NotationDeclHandler(self, *args):
            name, base, sysid, pubid = args
            self.out.append('Notation declared: %s' %(args,))

        def UnparsedEntityDeclHandler(self, *args):
            entityName, base, systemId, publicId, notationName = args
            self.out.append('Unparsed entity decl: %s' %(args,))

        def NotStandaloneHandler(self, userData):
            self.out.append('Not standalone')
            return 1

        def ExternalEntityRefHandler(self, *args):
            context, base, sysId, pubId = args
            self.out.append('External entity ref: %s' %(args[1:],))
            return 1

        def DefaultHandler(self, userData):
            pass

        def DefaultHandlerExpand(self, userData):
            pass

    handler_names = [
        'StartElementHandler', 'EndElementHandler',
        'CharacterDataHandler', 'ProcessingInstructionHandler',
        'UnparsedEntityDeclHandler', 'NotationDeclHandler',
        'StartNamespaceDeclHandler', 'EndNamespaceDeclHandler',
        'CommentHandler', 'StartCdataSectionHandler',
        'EndCdataSectionHandler',
        'DefaultHandler', 'DefaultHandlerExpand',
        #'NotStandaloneHandler',
        'ExternalEntityRefHandler'
        ]

    def test_utf8(self):

        out = self.Outputter()
        parser = expat.ParserCreate(namespace_separator='!')
        for name in self.handler_names:
            setattr(parser, name, getattr(out, name))
        parser.returns_unicode = 0
        parser.Parse(data, 1)

        # Verify output
        op = out.out
        self.assertEqual(op[0], 'PI: \'xml-stylesheet\' \'href="stylesheet.css"\'')
        self.assertEqual(op[1], "Comment: ' comment data '")
        self.assertEqual(op[2], "Notation declared: ('notation', None, 'notation.jpeg', None)")
        self.assertEqual(op[3], "Unparsed entity decl: ('unparsed_entity', None, 'entity.file', None, 'notation')")
        self.assertEqual(op[4], "Start element: 'root' {'attr1': 'value1', 'attr2': 'value2\\xe1\\xbd\\x80'}")
        self.assertEqual(op[5], "NS decl: 'myns' 'http://www.python.org/namespace'")
        self.assertEqual(op[6], "Start element: 'http://www.python.org/namespace!subelement' {}")
        self.assertEqual(op[7], "Character data: 'Contents of subelements'")
        self.assertEqual(op[8], "End element: 'http://www.python.org/namespace!subelement'")
        self.assertEqual(op[9], "End of NS decl: 'myns'")
        self.assertEqual(op[10], "Start element: 'sub2' {}")
        self.assertEqual(op[11], 'Start of CDATA section')
        self.assertEqual(op[12], "Character data: 'contents of CDATA section'")
        self.assertEqual(op[13], 'End of CDATA section')
        self.assertEqual(op[14], "End element: 'sub2'")
        self.assertEqual(op[15], "External entity ref: (None, 'entity.file', None)")
        self.assertEqual(op[16], "End element: 'root'")

    def test_unicode(self):
        # Try the parse again, this time producing Unicode output
        out = self.Outputter()
        parser = expat.ParserCreate(namespace_separator='!')
        parser.returns_unicode = 1
        for name in self.handler_names:
            setattr(parser, name, getattr(out, name))

        parser.Parse(data, 1)

        op = out.out
        self.assertEqual(op[0], 'PI: u\'xml-stylesheet\' u\'href="stylesheet.css"\'')
        self.assertEqual(op[1], "Comment: u' comment data '")
        self.assertEqual(op[2], "Notation declared: (u'notation', None, u'notation.jpeg', None)")
        self.assertEqual(op[3], "Unparsed entity decl: (u'unparsed_entity', None, u'entity.file', None, u'notation')")
        self.assertEqual(op[4], "Start element: u'root' {u'attr1': u'value1', u'attr2': u'value2\\u1f40'}")
        self.assertEqual(op[5], "NS decl: u'myns' u'http://www.python.org/namespace'")
        self.assertEqual(op[6], "Start element: u'http://www.python.org/namespace!subelement' {}")
        self.assertEqual(op[7], "Character data: u'Contents of subelements'")
        self.assertEqual(op[8], "End element: u'http://www.python.org/namespace!subelement'")
        self.assertEqual(op[9], "End of NS decl: u'myns'")
        self.assertEqual(op[10], "Start element: u'sub2' {}")
        self.assertEqual(op[11], 'Start of CDATA section')
        self.assertEqual(op[12], "Character data: u'contents of CDATA section'")
        self.assertEqual(op[13], 'End of CDATA section')
        self.assertEqual(op[14], "End element: u'sub2'")
        self.assertEqual(op[15], "External entity ref: (None, u'entity.file', None)")
        self.assertEqual(op[16], "End element: u'root'")

    def test_parse_file(self):
        # Try parsing a file
        out = self.Outputter()
        parser = expat.ParserCreate(namespace_separator='!')
        parser.returns_unicode = 1
        for name in self.handler_names:
            setattr(parser, name, getattr(out, name))
        file = StringIO.StringIO(data)

        parser.ParseFile(file)

        op = out.out
        self.assertEqual(op[0], 'PI: u\'xml-stylesheet\' u\'href="stylesheet.css"\'')
        self.assertEqual(op[1], "Comment: u' comment data '")
        self.assertEqual(op[2], "Notation declared: (u'notation', None, u'notation.jpeg', None)")
        self.assertEqual(op[3], "Unparsed entity decl: (u'unparsed_entity', None, u'entity.file', None, u'notation')")
        self.assertEqual(op[4], "Start element: u'root' {u'attr1': u'value1', u'attr2': u'value2\\u1f40'}")
        self.assertEqual(op[5], "NS decl: u'myns' u'http://www.python.org/namespace'")
        self.assertEqual(op[6], "Start element: u'http://www.python.org/namespace!subelement' {}")
        self.assertEqual(op[7], "Character data: u'Contents of subelements'")
        self.assertEqual(op[8], "End element: u'http://www.python.org/namespace!subelement'")
        self.assertEqual(op[9], "End of NS decl: u'myns'")
        self.assertEqual(op[10], "Start element: u'sub2' {}")
        self.assertEqual(op[11], 'Start of CDATA section')
        self.assertEqual(op[12], "Character data: u'contents of CDATA section'")
        self.assertEqual(op[13], 'End of CDATA section')
        self.assertEqual(op[14], "End element: u'sub2'")
        self.assertEqual(op[15], "External entity ref: (None, u'entity.file', None)")
        self.assertEqual(op[16], "End element: u'root'")

        # Issue 4877: expat.ParseFile causes segfault on a closed file.
        fp = open(test_support.TESTFN, 'wb')
        try:
            fp.close()
            parser = expat.ParserCreate()
            with self.assertRaises(ValueError):
                parser.ParseFile(fp)
        finally:
            test_support.unlink(test_support.TESTFN)

    def test_parse_again(self):
        parser = expat.ParserCreate()
        file = StringIO.StringIO(data)
        parser.ParseFile(file)
        # Issue 6676: ensure a meaningful exception is raised when attempting
        # to parse more than one XML document per xmlparser instance,
        # a limitation of the Expat library.
        with self.assertRaises(expat.error) as cm:
            parser.ParseFile(file)
        self.assertEqual(expat.ErrorString(cm.exception.code),
                          expat.errors.XML_ERROR_FINISHED)

class NamespaceSeparatorTest(unittest.TestCase):
    def test_legal(self):
        # Tests that make sure we get errors when the namespace_separator value
        # is illegal, and that we don't for good values:
        expat.ParserCreate()
        expat.ParserCreate(namespace_separator=None)
        expat.ParserCreate(namespace_separator=' ')

    def test_illegal(self):
        try:
            expat.ParserCreate(namespace_separator=42)
            self.fail()
        except TypeError, e:
            self.assertEqual(str(e),
                'ParserCreate() argument 2 must be string or None, not int')

        try:
            expat.ParserCreate(namespace_separator='too long')
            self.fail()
        except ValueError, e:
            self.assertEqual(str(e),
                'namespace_separator must be at most one character, omitted, or None')

    def test_zero_length(self):
        # ParserCreate() needs to accept a namespace_separator of zero length
        # to satisfy the requirements of RDF applications that are required
        # to simply glue together the namespace URI and the localname.  Though
        # considered a wart of the RDF specifications, it needs to be supported.
        #
        # See XML-SIG mailing list thread starting with
        # http://mail.python.org/pipermail/xml-sig/2001-April/005202.html
        #
        expat.ParserCreate(namespace_separator='') # too short


class InterningTest(unittest.TestCase):
    def test(self):
        # Test the interning machinery.
        p = expat.ParserCreate()
        L = []
        def collector(name, *args):
            L.append(name)
        p.StartElementHandler = collector
        p.EndElementHandler = collector
        p.Parse("<e> <e/> <e></e> </e>", 1)
        tag = L[0]
        self.assertEqual(len(L), 6)
        for entry in L:
            # L should have the same string repeated over and over.
            self.assertTrue(tag is entry)


class BufferTextTest(unittest.TestCase):
    def setUp(self):
        self.stuff = []
        self.parser = expat.ParserCreate()
        self.parser.buffer_text = 1
        self.parser.CharacterDataHandler = self.CharacterDataHandler

    def check(self, expected, label):
        self.assertEqual(self.stuff, expected,
                "%s\nstuff    = %r\nexpected = %r"
                % (label, self.stuff, map(unicode, expected)))

    def CharacterDataHandler(self, text):
        self.stuff.append(text)

    def StartElementHandler(self, name, attrs):
        self.stuff.append("<%s>" % name)
        bt = attrs.get("buffer-text")
        if bt == "yes":
            self.parser.buffer_text = 1
        elif bt == "no":
            self.parser.buffer_text = 0

    def EndElementHandler(self, name):
        self.stuff.append("</%s>" % name)

    def CommentHandler(self, data):
        self.stuff.append("<!--%s-->" % data)

    def setHandlers(self, handlers=[]):
        for name in handlers:
            setattr(self.parser, name, getattr(self, name))

    def test_default_to_disabled(self):
        parser = expat.ParserCreate()
        self.assertFalse(parser.buffer_text)

    def test_buffering_enabled(self):
        # Make sure buffering is turned on
        self.assertTrue(self.parser.buffer_text)
        self.parser.Parse("<a>1<b/>2<c/>3</a>", 1)
        self.assertEqual(self.stuff, ['123'],
                         "buffered text not properly collapsed")

    def test1(self):
        # XXX This test exposes more detail of Expat's text chunking than we
        # XXX like, but it tests what we need to concisely.
        self.setHandlers(["StartElementHandler"])
        self.parser.Parse("<a>1<b buffer-text='no'/>2\n3<c buffer-text='yes'/>4\n5</a>", 1)
        self.assertEqual(self.stuff,
                         ["<a>", "1", "<b>", "2", "\n", "3", "<c>", "4\n5"],
                         "buffering control not reacting as expected")

    def test2(self):
        self.parser.Parse("<a>1<b/>&lt;2&gt;<c/>&#32;\n&#x20;3</a>", 1)
        self.assertEqual(self.stuff, ["1<2> \n 3"],
                         "buffered text not properly collapsed")

    def test3(self):
        self.setHandlers(["StartElementHandler"])
        self.parser.Parse("<a>1<b/>2<c/>3</a>", 1)
        self.assertEqual(self.stuff, ["<a>", "1", "<b>", "2", "<c>", "3"],
                          "buffered text not properly split")

    def test4(self):
        self.setHandlers(["StartElementHandler", "EndElementHandler"])
        self.parser.CharacterDataHandler = None
        self.parser.Parse("<a>1<b/>2<c/>3</a>", 1)
        self.assertEqual(self.stuff,
                         ["<a>", "<b>", "</b>", "<c>", "</c>", "</a>"])

    def test5(self):
        self.setHandlers(["StartElementHandler", "EndElementHandler"])
        self.parser.Parse("<a>1<b></b>2<c/>3</a>", 1)
        self.assertEqual(self.stuff,
            ["<a>", "1", "<b>", "</b>", "2", "<c>", "</c>", "3", "</a>"])

    def test6(self):
        self.setHandlers(["CommentHandler", "EndElementHandler",
                    "StartElementHandler"])
        self.parser.Parse("<a>1<b/>2<c></c>345</a> ", 1)
        self.assertEqual(self.stuff,
            ["<a>", "1", "<b>", "</b>", "2", "<c>", "</c>", "345", "</a>"],
            "buffered text not properly split")

    def test7(self):
        self.setHandlers(["CommentHandler", "EndElementHandler",
                    "StartElementHandler"])
        self.parser.Parse("<a>1<b/>2<c></c>3<!--abc-->4<!--def-->5</a> ", 1)
        self.assertEqual(self.stuff,
                         ["<a>", "1", "<b>", "</b>", "2", "<c>", "</c>", "3",
                          "<!--abc-->", "4", "<!--def-->", "5", "</a>"],
                         "buffered text not properly split")


# Test handling of exception from callback:
class HandlerExceptionTest(unittest.TestCase):
    def StartElementHandler(self, name, attrs):
        raise RuntimeError(name)

    def test(self):
        parser = expat.ParserCreate()
        parser.StartElementHandler = self.StartElementHandler
        try:
            parser.Parse("<a><b><c/></b></a>", 1)
            self.fail()
        except RuntimeError, e:
            self.assertEqual(e.args[0], 'a',
                             "Expected RuntimeError for element 'a', but" + \
                             " found %r" % e.args[0])


# Test Current* members:
class PositionTest(unittest.TestCase):
    def StartElementHandler(self, name, attrs):
        self.check_pos('s')

    def EndElementHandler(self, name):
        self.check_pos('e')

    def check_pos(self, event):
        pos = (event,
               self.parser.CurrentByteIndex,
               self.parser.CurrentLineNumber,
               self.parser.CurrentColumnNumber)
        self.assertTrue(self.upto < len(self.expected_list),
                        'too many parser events')
        expected = self.expected_list[self.upto]
        self.assertEqual(pos, expected,
                'Expected position %s, got position %s' %(pos, expected))
        self.upto += 1

    def test(self):
        self.parser = expat.ParserCreate()
        self.parser.StartElementHandler = self.StartElementHandler
        self.parser.EndElementHandler = self.EndElementHandler
        self.upto = 0
        self.expected_list = [('s', 0, 1, 0), ('s', 5, 2, 1), ('s', 11, 3, 2),
                              ('e', 15, 3, 6), ('e', 17, 4, 1), ('e', 22, 5, 0)]

        xml = '<a>\n <b>\n  <c/>\n </b>\n</a>'
        self.parser.Parse(xml, 1)


class sf1296433Test(unittest.TestCase):
    def test_parse_only_xml_data(self):
        # http://python.org/sf/1296433
        #
        xml = "<?xml version='1.0' encoding='iso8859'?><s>%s</s>" % ('a' * 1025)
        # this one doesn't crash
        #xml = "<?xml version='1.0'?><s>%s</s>" % ('a' * 10000)

        class SpecificException(Exception):
            pass

        def handler(text):
            raise SpecificException

        parser = expat.ParserCreate()
        parser.CharacterDataHandler = handler

        self.assertRaises(Exception, parser.Parse, xml)

class ChardataBufferTest(unittest.TestCase):
    """
    test setting of chardata buffer size
    """

    def test_1025_bytes(self):
        self.assertEqual(self.small_buffer_test(1025), 2)

    def test_1000_bytes(self):
        self.assertEqual(self.small_buffer_test(1000), 1)

    def test_wrong_size(self):
        parser = expat.ParserCreate()
        parser.buffer_text = 1
        def f(size):
            parser.buffer_size = size

        self.assertRaises(TypeError, f, sys.maxint+1)
        self.assertRaises(ValueError, f, -1)
        self.assertRaises(ValueError, f, 0)

    def test_unchanged_size(self):
        xml1 = ("<?xml version='1.0' encoding='iso8859'?><s>%s" % ('a' * 512))
        xml2 = 'a'*512 + '</s>'
        parser = expat.ParserCreate()
        parser.CharacterDataHandler = self.counting_handler
        parser.buffer_size = 512
        parser.buffer_text = 1

        # Feed 512 bytes of character data: the handler should be called
        # once.
        self.n = 0
        parser.Parse(xml1)
        self.assertEqual(self.n, 1)

        # Reassign to buffer_size, but assign the same size.
        parser.buffer_size = parser.buffer_size
        self.assertEqual(self.n, 1)

        # Try parsing rest of the document
        parser.Parse(xml2)
        self.assertEqual(self.n, 2)


    def test_disabling_buffer(self):
        xml1 = "<?xml version='1.0' encoding='iso8859'?><a>%s" % ('a' * 512)
        xml2 = ('b' * 1024)
        xml3 = "%s</a>" % ('c' * 1024)
        parser = expat.ParserCreate()
        parser.CharacterDataHandler = self.counting_handler
        parser.buffer_text = 1
        parser.buffer_size = 1024
        self.assertEqual(parser.buffer_size, 1024)

        # Parse one chunk of XML
        self.n = 0
        parser.Parse(xml1, 0)
        self.assertEqual(parser.buffer_size, 1024)
        self.assertEqual(self.n, 1)

        # Turn off buffering and parse the next chunk.
        parser.buffer_text = 0
        self.assertFalse(parser.buffer_text)
        self.assertEqual(parser.buffer_size, 1024)
        for i in range(10):
            parser.Parse(xml2, 0)
        self.assertEqual(self.n, 11)

        parser.buffer_text = 1
        self.assertTrue(parser.buffer_text)
        self.assertEqual(parser.buffer_size, 1024)
        parser.Parse(xml3, 1)
        self.assertEqual(self.n, 12)



    def make_document(self, bytes):
        return ("<?xml version='1.0'?><tag>" + bytes * 'a' + '</tag>')

    def counting_handler(self, text):
        self.n += 1

    def small_buffer_test(self, buffer_len):
        xml = "<?xml version='1.0' encoding='iso8859'?><s>%s</s>" % ('a' * buffer_len)
        parser = expat.ParserCreate()
        parser.CharacterDataHandler = self.counting_handler
        parser.buffer_size = 1024
        parser.buffer_text = 1

        self.n = 0
        parser.Parse(xml)
        return self.n

    def test_change_size_1(self):
        xml1 = "<?xml version='1.0' encoding='iso8859'?><a><s>%s" % ('a' * 1024)
        xml2 = "aaa</s><s>%s</s></a>" % ('a' * 1025)
        parser = expat.ParserCreate()
        parser.CharacterDataHandler = self.counting_handler
        parser.buffer_text = 1
        parser.buffer_size = 1024
        self.assertEqual(parser.buffer_size, 1024)

        self.n = 0
        parser.Parse(xml1, 0)
        parser.buffer_size *= 2
        self.assertEqual(parser.buffer_size, 2048)
        parser.Parse(xml2, 1)
        self.assertEqual(self.n, 2)

    def test_change_size_2(self):
        xml1 = "<?xml version='1.0' encoding='iso8859'?><a>a<s>%s" % ('a' * 1023)
        xml2 = "aaa</s><s>%s</s></a>" % ('a' * 1025)
        parser = expat.ParserCreate()
        parser.CharacterDataHandler = self.counting_handler
        parser.buffer_text = 1
        parser.buffer_size = 2048
        self.assertEqual(parser.buffer_size, 2048)

        self.n=0
        parser.Parse(xml1, 0)
        parser.buffer_size //= 2
        self.assertEqual(parser.buffer_size, 1024)
        parser.Parse(xml2, 1)
        self.assertEqual(self.n, 4)

class MalformedInputText(unittest.TestCase):
    def test1(self):
        xml = "\0\r\n"
        parser = expat.ParserCreate()
        try:
            parser.Parse(xml, True)
            self.fail()
        except expat.ExpatError as e:
            self.assertEqual(str(e), 'unclosed token: line 2, column 0')

    def test2(self):
        xml = "<?xml version\xc2\x85='1.0'?>\r\n"
        parser = expat.ParserCreate()
        try:
            parser.Parse(xml, True)
            self.fail()
        except expat.ExpatError as e:
            self.assertEqual(str(e), 'XML declaration not well-formed: line 1, column 14')

class ForeignDTDTests(unittest.TestCase):
    """
    Tests for the UseForeignDTD method of expat parser objects.
    """
    def test_use_foreign_dtd(self):
        """
        If UseForeignDTD is passed True and a document without an external
        entity reference is parsed, ExternalEntityRefHandler is first called
        with None for the public and system ids.
        """
        handler_call_args = []
        def resolve_entity(context, base, system_id, public_id):
            handler_call_args.append((public_id, system_id))
            return 1

        parser = expat.ParserCreate()
        parser.UseForeignDTD(True)
        parser.SetParamEntityParsing(expat.XML_PARAM_ENTITY_PARSING_ALWAYS)
        parser.ExternalEntityRefHandler = resolve_entity
        parser.Parse("<?xml version='1.0'?><element/>")
        self.assertEqual(handler_call_args, [(None, None)])

        # test UseForeignDTD() is equal to UseForeignDTD(True)
        handler_call_args[:] = []

        parser = expat.ParserCreate()
        parser.UseForeignDTD()
        parser.SetParamEntityParsing(expat.XML_PARAM_ENTITY_PARSING_ALWAYS)
        parser.ExternalEntityRefHandler = resolve_entity
        parser.Parse("<?xml version='1.0'?><element/>")
        self.assertEqual(handler_call_args, [(None, None)])

    def test_ignore_use_foreign_dtd(self):
        """
        If UseForeignDTD is passed True and a document with an external
        entity reference is parsed, ExternalEntityRefHandler is called with
        the public and system ids from the document.
        """
        handler_call_args = []
        def resolve_entity(context, base, system_id, public_id):
            handler_call_args.append((public_id, system_id))
            return 1

        parser = expat.ParserCreate()
        parser.UseForeignDTD(True)
        parser.SetParamEntityParsing(expat.XML_PARAM_ENTITY_PARSING_ALWAYS)
        parser.ExternalEntityRefHandler = resolve_entity
        parser.Parse(
            "<?xml version='1.0'?><!DOCTYPE foo PUBLIC 'bar' 'baz'><element/>")
        self.assertEqual(handler_call_args, [("bar", "baz")])


def test_main():
    run_unittest(SetAttributeTest,
                 ParseTest,
                 NamespaceSeparatorTest,
                 InterningTest,
                 BufferTextTest,
                 HandlerExceptionTest,
                 PositionTest,
                 sf1296433Test,
                 ChardataBufferTest,
                 MalformedInputText,
                 ForeignDTDTests)

if __name__ == "__main__":
    test_main()



import subprocess, shlex
import whatap.util.thread_util as thread_util
import whatap.util.logging_util as logging_util
import whatap.util.process_util as process_util
import whatap.pack.smbase_pack as smbase_pack
import whatap.agent.conf.configure as conf
import time, os
import psutil
import psutil._psutil_posix
from . import CPUStat


class _vmstat:
    def __init__(self):
        self.sys = 0
        self.user = 0
        self.idle = 0
        self.wait = 0



initVmstatFlag = False
initPollVmstatFlag = False
infraVmstat = _vmstat()
serverVmstat = _vmstat()

# polling process 
def pollVmstatForInfra(vmInterval = 2):
    global infraVmstat
    command = "vmstat -I {} 30".format(vmInterval)
    #command = "vmstat -I {} 1800".format(vmInterval)
    process = process_util.pollCommand(command)

    while True:
        line = process.stdout.readline()
        if line:
            fields = line.split()
            if len(fields) < 18 or fields[0] == "kthr" or fields[0].startswith("---") or fields[0] == "r":
                logging_util.debug(fields)
                continue
            vm = _vmstat()
            vm.user = float(fields[14])
            vm.sys = float(fields[15])
            vm.idle = float(fields[16])
            vm.wait = float(fields[17])
            infraVmstat = vm

        else:
            if process.poll() is not None:
                break

    process.wait()


def pollVmstatForServer(vmInterval = 5):
    global serverVmstat
    command = "vmstat -I {} 12".format(vmInterval)
    #command = "vmstat -I {} 720".format(vmInterval)
    process = process_util.pollCommand(command)

    while True:
        line = process.stdout.readline()
        if line:
            fields = line.split()
            if len(fields) < 18 or fields[0] == "kthr" or fields[0].startswith("---") or fields[0] == "r":
                logging_util.debug(fields)
                continue
            vm = _vmstat()
            vm.user = float(fields[14])
            vm.sys = float(fields[15])
            vm.idle = float(fields[16])
            vm.wait = float(fields[17])
            serverVmstat = vm


        else:
            if process.poll() is not None:
                break

    process.wait()



def pollVmstat():
    infraThread = None
    serverThread = None
    while True:
        if conf.GetConfig().CategoryV2Enabled and (infraThread == None or not infraThread.is_alive()):
            infraThread = thread_util.async(pollVmstatForInfra, 2)
        if serverThread == None or not serverThread.is_alive():
            serverThread = thread_util.async(pollVmstatForServer, 5)
        
        time.sleep(1)

def initPollVmstat():
    global initPollVmstatFlag
    if not initPollVmstatFlag:
        thread_util.async(pollVmstat)
        initPollVmstatFlag = True


# non polling process 
def runVmstatForInfra(vmInterval = 2):
    global infraVmstat
    command = "vmstat -I {} 1".format(vmInterval)
    output = process_util.executeCommandShellWithTimeout(command, 5)
    vmstat = output.split("\n")[-1]
    vmstat = vmstat.split()

    if len(vmstat) < 18:
        logging_util.error("vmstat error command : {}, result : {}".format(command, output))
        return None

    vm = _vmstat()
    vm.user = float(vmstat[14])
    vm.sys = float(vmstat[15])
    vm.idle = float(vmstat[16])
    vm.wait = float(vmstat[17])
    infraVmstat = vm

    return vmstat


def runVmstatForServer(vmInterval = 5):
    global serverVmstat
    command = "vmstat -I {} 1".format(vmInterval)
    output = process_util.executeCommandShellWithTimeout(command, 10)
    vmstat = output.split("\n")[-1]
    vmstat = vmstat.split()

    if len(vmstat) < 18:
        logging_util.error("vmstat error command : {}, result : {}".format(command, output))
        return None

    vm = _vmstat()
    vm.user = float(vmstat[14])
    vm.sys = float(vmstat[15])
    vm.idle = float(vmstat[16])
    vm.wait = float(vmstat[17])
    serverVmstat = vm

    return vmstat

def initRunVmstat():
    global initVmstatFlag
    if not initVmstatFlag:
        try:
            if conf.GetConfig().CategoryV2Enabled:
                thread_util.asyncLoop(runVmstatForInfra, 2)
            thread_util.asyncLoop(runVmstatForServer, 5)
        except Exception, e:
            logging_util.error(e)
    
        initVmstatFlag = True



oldCpu = None
# infra
def GetCpu():
    #initRunVmstat()
    initPollVmstat
    global infraVmstat
    cpu_times = psutil.cpu_times(percpu=False)

    global oldCpu
    if not oldCpu:
        oldCpu = cpu_times
        return None

    c = smbase_pack.CpuLinux()
    #all_cpu = cpu_times.user + cpu_times.iowait + cpu_times.idle + cpu_times.system
    #last_all_cpu = oldCpu.user + oldCpu.iowait + oldCpu.idle + oldCpu.system
    jiff = float(100)
    #cpu_diff = float(all_cpu - last_all_cpu)
    #c.user = jiff * float(cpu_times.user - oldCpu.user) / cpu_diff
    #c.system = jiff * float(cpu_times.system - oldCpu.system) / cpu_diff
    #c.iowait = jiff * float(cpu_times.iowait - oldCpu.iowait) / cpu_diff
    #c.idle = jiff * float(cpu_times.idle - oldCpu.idle) / cpu_diff
    c.user = infraVmstat.user
    c.system = infraVmstat.sys
    c.idle = infraVmstat.idle
    c.iowait = infraVmstat.wait

    c.nice = float(0)
    c.irq = float(0)
    c.softirq = float(0)
    c.steal = float(0)
    c.load1, c.load5, c.load15= psutil.cpu_load()
    oldCpu = cpu_times

    return c

oldCpuList = None
def GetCpuCore():
    cpu_times = psutil.logical_cpu_times(percpu=True)
    jiff = float(100)

    global oldCpuList
    if not oldCpuList:
        oldCpuList = cpu_times
        return None

    cpuPerf = []
    for index, ctimes in enumerate(cpu_times):
        c = smbase_pack.CpuLinux()
        all_cpu = ctimes.user + ctimes.iowait + ctimes.idle + ctimes.system
        last_all_cpu = oldCpuList[index].user + oldCpuList[index].iowait + oldCpuList[index].idle + oldCpuList[index].system
        cpu_diff = float(all_cpu - last_all_cpu)

        if cpu_diff == 0:
            c.user = 0
            c.system = 0
            c.iowait = 0
            c.idle = 0
        else:
            c.user = jiff * float(ctimes.user - oldCpuList[index].user) / cpu_diff
            c.system = jiff * float(ctimes.system - oldCpuList[index].system) / cpu_diff
            c.iowait = jiff * float(ctimes.iowait - oldCpuList[index].iowait) / cpu_diff
            c.idle = jiff * float(ctimes.idle - oldCpuList[index].idle) / cpu_diff

        c.nice = float(0)
        c.irq = float(0)
        c.softirq = float(0)
        c.steal = float(0)

        cpuPerf.append(c) 

    oldCpuList = cpu_times

    return cpuPerf


oldStats = None
def GetCpuStat():
    now = time.time()

    cpuStats = psutil.cpu_stats_detail()

    global oldStats
    if not oldStats:
        oldStats = [cpuStats, now]
        return None

    difftime = float(now - oldStats[1])
    oldCpuStats = oldStats[0]

    cs = CPUStat() 
    cs.ctx = float(cpuStats[0] - oldCpuStats[0]) / difftime
    cs.interrupt = float(cpuStats[1] - oldCpuStats[1]) / difftime
    cs.syscall = float(cpuStats[2] - oldCpuStats[2]) / difftime
    cs.forkCount = float(cpuStats[3] - oldCpuStats[3]) / difftime
    cs.execCount = float(cpuStats[4] - oldCpuStats[4]) / difftime
    cs.runqueue = float(cpuStats[5] - oldCpuStats[5]) / difftime
    cs.blockqueue = float(cpuStats[6] - oldCpuStats[6]) / difftime
    cs.pendingqueue = float(cpuStats[7] - oldCpuStats[7]) / difftime
    cs.waitqueue = CPUStat.UNSUPPORTED

    oldStats = [cpuStats, now]

    return cs

def GetSwapMemory():
    m = smbase_pack.MemoryLinux()
    swapmem = psutil.swap_memory()
    m.swapUsed = swapmem.used
    m.swapTotal = swapmem.total

    if m.swapTotal > 0:
        m.swapPused = swapmem.percent
    else:
        m.swapPused = 0

    return m


oldMemoryStat = None
def GetMemory():
    memory = psutil.virtual_memory_detail()
    now = time.time()

    global oldMemoryStat
    if not oldMemoryStat:
        oldMemoryStat = [memory, now]
        return None


    m = smbase_pack.MemoryLinux()
    m.total = memory[0] # total
    m.free = memory[2] # free
    m.cached = memory[5] #numperm
    m.available = m.free+ m.cached
    m.pavailable = float(m.available) / float(m.total) * float(100.0)
    m.used = m.total - m.available
    m.pused = float(float(m.used)/float(m.total)*100.0)
    m.buffers = 0
    m.shared = 0
    m.pageFault = 0
 
    oldMemory = oldMemoryStat[0]
    difftime = float(now - oldMemoryStat[1])

    m.swapIn = float(memory[6] - oldMemory[6]) / difftime
    m.swapOut = float(memory[7] - oldMemory[7]) / difftime
    m.fileSystemIn = float(memory[8] - oldMemory[8]) / difftime
    m.fileSystemOut = float(memory[9] - oldMemory[9]) / difftime
    m.scannedPage = float(memory[10] - oldMemory[10]) / difftime
    m.freedPage = float(memory[11] - oldMemory[11]) / difftime

    oldMemoryStat = [memory, now]
    return m

oldVmstat = None
def GetCpuMemory():
    #initRunVmstat()
    initPollVmstat()
    global serverVmstat

    cpu_times = psutil.cpu_times(percpu=False)
    global oldVmstat
    if not oldVmstat:
        oldVmstat = cpu_times
        return None, None, None

    c = smbase_pack.CpuLinux()
    #all_cpu = cpu_times.user + cpu_times.iowait + cpu_times.idle + cpu_times.system
    #last_all_cpu = oldVmstat.user + oldVmstat.iowait + oldVmstat.idle + oldVmstat.system
    #cpu_diff = float(all_cpu - last_all_cpu)
    
    c.user = serverVmstat.user
    c.system = serverVmstat.sys
    c.idle = serverVmstat.idle
    c.iowait = serverVmstat.wait
    #c.user = float(100) * float(cpu_times.user - oldVmstat.user) / cpu_diff
    #c.system = float(100) * float(cpu_times.system - oldVmstat.system) / cpu_diff
    #c.iowait = float(100) * float(cpu_times.iowait - oldVmstat.iowait) / cpu_diff
    #c.idle = float(100) * float(cpu_times.idle - oldVmstat.idle) / cpu_diff
    c.nice = float(0)
    c.irq = float(0)
    c.softirq = float(0)
    c.steal = float(0)
    c.load1, c.load5, c.load15= psutil.cpu_load()
    oldVmstat = cpu_times

    

    p = smbase_pack.MemoryLinux()
    memstat = getMemStat()
    p.total = memstat['total']
    p.free = memstat['free']  
    p.cached = memstat['cached']
    p.available = p.total - p.cached- p.free
    p.pavailable = float(p.available) / float(p.total) * float(100.0)
    p.used = memstat['used']
    p.pused = float(p.used) / float(p.total) * float(100.0)
       
    p.buffers = 0
    p.shared = 0

    swapmem = psutil.swap_memory()
    p.swapUsed = swapmem.used
    p.swapTotal = swapmem.total
    if p.swapTotal > 0:
        p.swapPused = swapmem.percent
    else:
        p.swapPused = 0

    p.pageFault = 0

    uptime= int(time.time()) - int(psutil.boot_time())
    #print 'cpu_memory', int(time.time()) , int(vmstat['boot time']/1024.0), uptime
    return c, p, uptime

g_memstat = {"total":1,"cached":0,"used":0,"free":0}

def getMemStat():
    global g_memstat

    return g_memstat


import svmon_parser
def getSVMon(pagesize = psutil._psutil_posix.getpagesize(), command="svmon -G"):
    started=time.time()*1000
    process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE)
    
    lines = ""
    while True:
        output = process.stdout.readline()
        #print 'getVmStat', output
        if not output and process.poll() is not None:
            break
        if output:
            lines += output

    svmon_output = svmon_parser.parse_svmon_output(lines)
    memory_paragraph = svmon_output["memory"]
    pininuse_paragraph = svmon_output["pin_in_use"]
    inuse_paragraph = svmon_output["in_use"]
    mem_total = memory_paragraph["size"] * pagesize
    mem_free = memory_paragraph["free"] * pagesize
    mem_cached = inuse_paragraph['clnt'] * pagesize
    mem_used = inuse_paragraph['inuse'] * pagesize

    return {"total":mem_total,"cached":mem_cached,"used":mem_used,"free":mem_free}

def updateMemstat():
    global g_memstat
    global memstat_keys
    
    memstat_thistime = getSVMon()
    
    g_memstat = memstat_thistime

    return g_memstat


try:
    updateMemstat()
    thread_util.asyncLoop(updateMemstat, conf.GetConfig().MemstatInterval)
except Exception, e:
    print(e)





import zlib
import os
import hash_util as HashUtil

#TEST
home = "./"

def get_crc(file_path):
    crc = 0
    print "a"
    try:
        with open(file_path, "rb") as f:
            buf = f.read()
            print buf
            crc =  HashUtil.hash(buf)
            print crc
    except Exception, e:
        print e
        return 0
    return crc & 0xffffffff

def delete(file_name):
    pass


def save(file_name, pos, data):
    tmp_file = os.path.join(home, file_name + ".tmp")

    if pos == 0:
        if os.path.exists(tmp_file):
            try:
                os.remove(tmp_file)
            except Exception:
                print "error1"
    else:
        if not os.path.exists(tmp_file) or os.path.getsize(tmp_file) < pos:
            return 
    mode = "w+b" if pos == 0 else "r+b"
    try:
        with open(tmp_file, mode) as raf:
            raf.seek(pos)
            raf.write(data)
        print "ok"
    except Exception:
        print "error2"

def done(file_name, size, crc32):
    tmp_file = os.path.join(home, file_name + ".tmp")
    if not os.path.exists(tmp_file) or os.path.getsize(tmp_file) != size:
        return 

    fcrc = get_crc(tmp_file)
    if fcrc == 0 or fcrc != crc32:
        return 

    final_file = os.path.join(home, file_name)
    try:
        os.rename(tmp_file, final_file)
    except Exception:
        pass



if __name__ == "__main__":
    test_file = "whatap_infra"
    test_data = b"TEST, DATA"
    save(test_file, 0, test_data)

    tmp_path = os.path.join(home, test_file + ".tmp")
    if os.path.exists(tmp_path):
        file_size = os.path.getsize(tmp_path)
        file_crc = get_crc(tmp_path)
        print("Saved tmp file size:", file_size)
        print("Saved tmp file crc32:", file_crc)
        done(test_file, file_size, file_crc)
        final_path = os.path.join(home, test_file)
        if os.path.exists(final_path):
            print("Final file exists:", final_path)
                                                                                                
